/*
    This file is part of Reversi DS.

    Reversi DS is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Reversi DS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Reversi DS.  If not, see <http://www.gnu.org/licenses/>.
     
    http://ReversiDS.sourceforge.net
*/
#pragma once
#include <iostream>
#include <vector>
#include <algorithm>
using namespace std;

//Prime number for simplistic hash (look into better hash function)
#define MAXHASH 42433
template<class GAME, class SIG>
class Maxmin{
	/* Template assumptions:
	 * 	GAME:
	 * 		provides int evaluate(), 
	 * 					 int ended(), 
	 * 					 vector<GAME> enumerateOptions()
	 * 					 SIG getSig()
	 * 					 bool operator<()
	 * SIG:
	 * 		provides 	unsigned int getHash(int MAXHASH)
	 * 					bool equals(SIG)
	 */ 
	
	
	
	///return maximized option score (alpha beta algorithm)
	/// @param GAME& the game state to evaluate 
	/// @param int	depth of search
	/// @param int  alpha value
	/// @param int  beta value
	int max(const GAME &,int,int,int);
	///return minimized option score (alpha beta algorithm)
	/// @param GAME& the game state to evaluate 
	/// @param int	depth of search
	/// @param int  alpha value
	/// @param int  beta value
	int min(const GAME &,int,int,int);
	///update hash table with new value
	/// @param GAME& new entry
	/// @param int the new value
	/// @param int alpha value
	/// @param int beta value
	void setHash( const GAME &, const int, const int, const int );
	/// called in evaluate to invalidate previous hash table values
	/// 	used so that a new hash table doesn't need to be allocated 
	///		each run
	void refresh(){
		stamp++;
	}
	/// see if a state has been already evaluated
	///		needs to be the current iteration, current window (alpha, beta) range needs to be smaller (otherwise might rule out some solutions)
	///		and have a unique match (not just a hash hit )
	///	@param GAME& the game state
	/// @param int& returns the value of the state, valid only if the state was hashed
	/// @param const int	alpha value
	/// @param const int	beta value
	/// @return true if the state has been evaluated, false otherwise
	bool hashed( const GAME &, int &, const int, const int)const;
	/// marks the current iteration
	int stamp;
	class LookUp{
		public:
		SIG sig;
		int value;
		int stamp;
		int alpha;
		int beta;
	};
	LookUp *hashTable;
	
public:
	/// find 'best' play
	/// @param in	the starting gamestate
	/// @param out	the gamestate after the 'best' move ( must be separate variable from in ) 
	/// @param depth depth to search
	void evaluate(GAME &in,GAME &out,int depth); 
	Maxmin(){
		this->stamp = 0;
		hashTable = new LookUp[MAXHASH];
		for(int x=0;x<MAXHASH; x++)
			hashTable[x].stamp=-1;
	}
	~Maxmin(){
		delete [] hashTable;
	}	
	
};


template<class GAME, class SIG>
void Maxmin<GAME, SIG>::evaluate(GAME &in,GAME &out,int depth){
	refresh();
	vector<GAME> options=in.enumerateOptions();
	sort(options.begin(),options.end());
	int score;
	
	int alpha= -10000;
	int beta= 10000;
	for(typename vector<GAME>::iterator it=options.begin(); it!=options.end(); ++it){
		if( !hashed( *it, score, alpha, beta ) ){
			score=min( *it,depth,alpha,beta);
			setHash( *it, score, alpha, beta );
		}
			
		if(score>alpha){
			out=*it;
			alpha=score;
		}
	}
}
template<class GAME,class SIG>
int Maxmin<GAME,SIG>::max(const GAME &in,int depth,int alpha,int beta){
	if(depth==0 || in.ended()){
		int score = in.evaluate();
		return score;
	}
	depth--;
	vector<GAME> options=in.enumerateOptions();
	//check hashmap signatures
	for( typename vector<GAME>::iterator i=options.begin(); i!=options.end(); i++){
		int score;
		if( hashed( *i, score, alpha, beta ) ){
			if( score > alpha ){
				alpha = score;
				if( alpha >= beta )
					return alpha;
			}
		}
	}	
	if(depth!=0)
		sort(options.begin(),options.end());
	int score;
	for(typename vector<GAME>::iterator it=options.begin();it!=options.end();it++){
		if( !hashed( *it, score, alpha, beta ) ){
			score=min(*it,depth,alpha,beta);
			setHash( *it, score, alpha, beta );
		}
		if(score>alpha)
			alpha=score;
		//only need to check this if alpha changes?
		if(alpha>=beta)
			break;

	}	
	return alpha;
}

template<class GAME,class SIG>
int Maxmin<GAME,SIG>::min(const GAME &in,int depth,int alpha,int beta){
	if(depth==0 || in.ended()){
		int score = in.evaluate();
		return score;
	}
	depth--;
	vector<GAME> options=in.enumerateOptions();
	for( typename vector<GAME>::iterator i=options.begin(); i!=options.end(); i++){
		int score;
		if( hashed( *i, score, alpha, beta) ){
			if( score < beta ){
				beta = score;
				if( alpha >= beta )
					return beta;
			}
		}
	}
	if(depth!=0)
		sort(options.begin(),options.end());
	int score;
	for(typename vector<GAME>::iterator it=options.begin();it!=options.end();it++){
		if( !hashed(*it,score, alpha, beta) ){
			score=max(*it,depth,alpha,beta);
			setHash( *it, score, alpha, beta );
		}
		if(score<beta){
			beta=score;
		}
		if(alpha>=beta)
			break;
	}
	return beta;
}

template<class GAME,class SIG>
bool Maxmin<GAME, SIG>::hashed(const GAME &g, int &v, const int alpha,const int beta)const{
	SIG s = g.getSig();
	unsigned int hash = s.getHash(MAXHASH);
	const LookUp &lu =hashTable[hash];
	if( lu.stamp == stamp &&						// if current iteration 
		lu.alpha <= alpha && lu.beta >= beta &&		//	  current window is smaller/equal
		lu.sig.equals(s) ){							//	  and we have an actual match
			v = lu.value;
			return true;
	}
	return false;	
}

template<class GAME,class SIG>
void Maxmin<GAME, SIG>::setHash(const GAME &g,const int v, const int alpha,const int beta){
	SIG s = g.getSig();
	unsigned int hash = s.getHash(MAXHASH);
	LookUp &lu =hashTable[hash];
	if( lu.stamp != stamp ){
		lu.value = v;
		lu.sig = s;
		lu.stamp = stamp;
		lu.alpha = alpha;
		lu.beta = beta;
	} else if( alpha <= lu.alpha && beta >= lu.beta ){
		/// if there is a collision, we probably want the one with 
		/// the larger window
		lu.value = v;
		lu.sig = s;
		lu.alpha = alpha;
		lu.beta = beta;
	} 
}

